/*-
 * ~~~~~~licensing~~~~~~
 * resource-based-storage
 * ==========
 * Copyright (C) 2020 - 2025 EmDev LLC
 * ==========
 * You may not use this file except in accordance with the License Terms of the Copyright
 * Holder located at: https://entaxy.ru/eula . All copyrights, all intellectual property
 * rights to the Software and any copies are the property of the Copyright Holder. Unless
 * it is explicitly allowed the Copyright Holder, the User is prohibited from using the
 * Software for commercial purposes to provide services to third parties.
 * 
 * The Copyright Holder hereby declares that the Software is provided on an "AS IS".
 * Under no circumstances does the Copyright Holder guarantee or promise that the
 * Software provided by him will be suitable or not suitable for the specific purposes
 * of the User, that the Software will meet all commercial and personal subjective
 * expectations of the User, that the Software will work properly, without technical
 * errors, quickly and uninterruptedly.
 * 
 * Under no circumstances shall the Copyright Holder or its Affiliates is not liable
 * to the User for any direct or indirect losses of the User, his expenses or actual
 * damage, including, downtime; loss of bussines; lost profit; lost earnings; loss
 * or damage to data, property, etc.
 * ~~~~~~/licensing~~~~~~
 */
package ru.entaxy.platform.integration.applications.storage.resource.impl;

import java.util.Collection;
import java.util.HashMap;
import java.util.Map;
import java.util.Set;

public class ComponentCache<T extends AbstractResourceComponent<?>> implements Map<String, T> {

    final protected Map<String, T> cache = new HashMap<>();

    synchronized public T put(T value) {
        return cache.put(value.getComponentKey(), value);
    }

    public boolean has(T value) {
        return cache.containsKey(value.getComponentKey());
    }

    @Override
    public int size() {
        return cache.size();
    }

    @Override
    public boolean isEmpty() {
        return cache.isEmpty();
    }

    @Override
    public boolean containsKey(Object key) {
        return cache.containsKey(key);
    }

    @Override
    public boolean containsValue(Object value) {
        return cache.containsValue(value);
    }

    @Override
    public T get(Object key) {
        return cache.get(key);
    }

    @Override
    synchronized public T put(String key, T value) {
        return cache.put(key, value);
    }

    @Override
    synchronized public T remove(Object key) {
        return cache.remove(key);
    }

    @Override
    synchronized public void putAll(Map<? extends String, ? extends T> m) {
        cache.putAll(m);
    }

    @Override
    synchronized public void clear() {
        cache.clear();
    }

    @Override
    public Set<String> keySet() {
        return cache.keySet();
    }

    @Override
    public Collection<T> values() {
        return cache.values();
    }

    @Override
    public Set<Entry<String, T>> entrySet() {
        return cache.entrySet();
    }



}
